import x10.util.Random; 
import x10.util.Timer; 

public class Histogram {
    static val rand = new Random(System.nanoTime());

    private static abstract class Version {
	abstract def histogram(hist:Array[Int]{self.rank==1}, input:Array[Int]{self.rank==1}, numAsyncs:Int):void;
	abstract def description():String;

	def oracle(histo:Array[Int]{self.rank==1}, input:Array[Int]{self.rank==1}) {
	    for (i in input) {
		val value: int = input(i);
		val x:Int = value % histo.size;
		histo(x) += 1;
	    }
	}

	def check(hist:Array[Int]{self.rank==1}, input:Array[Int]{self.rank==1}) {
	    var correct:Array[Int]{self.rank==1} = new Array[Int](hist.size, (p:Int) => 0);
	    oracle(correct,input);
	    for (var i:Int = 0; i < hist.size; i++)
		if (hist(i) != correct(i))
		    throw new Exception("Incorrect result!");	    
	}
    }

    public static def pretty(x:Float):String{
	return String.format("%.2f", new Array[Any](1, x));
    }
    
    public static def test(v:Version, inputSize:Int, histoSize:Int, maxValue:Int, maxAsyncs:Int, numTrials:Int) {
	Console.OUT.println("============== " + v.description() + " ==============");
	Console.OUT.println("asyncs\tspeedup\ttime\ttimes");
	var base:Float = 0; // Deliberately trigger a divide by 0 if baseline not overwritten
	for (var n:Int = 1; n <= maxAsyncs; n = n*2) {

	    val numAsyncs = n;
	    val times = new Array[Long](numTrials,(p:Int) => {
		    // generate new input 
		    val input = new Array[Int](inputSize, (p:Int) => rand.nextInt(maxValue));
		    val histo = new Array[Int](histoSize, (p:Int) => 0);
		    val start = Timer.milliTime();
		    v.histogram(histo,input,numAsyncs);
		    val end = Timer.milliTime();
		    (end-start)
		});

	    val sum = times.reduce((a:Long,b:Long)=>a+b,0L);
	    val avg = sum/times.size;
 	    if (n == 1) base = avg;
 	    val speedup = base/avg;
 	    Console.OUT.println(n + "\t" + pretty(speedup) + "\t" + pretty(avg) + "\t" + times);		
	}
    }

    private static class Serial extends Version {
	def description():String {
	    return "Serial, idiomatic";
	}
	
	def histogram(hist:Array[Int]{self.rank==1}, input:Array[Int]{self.rank==1}, numAsyncs:Int):void {
	    oracle(hist,input);
	}
    }

    public static def main(argv:Array[String]{self.rank==1}) {
	if (argv.size != 5) {
	    Console.ERR.println("USAGE: Histogram <inputSize> <histoSize> <maxValue> <maxAsyncs> <numTrials>");
	    return;
	}
	val inputSize = Int.parseInt(argv(0));
	val histoSize = Int.parseInt(argv(1));
	val maxValue = Int.parseInt(argv(2));
	val maxAsyncs = Int.parseInt(argv(3));
	val numTrials = Int.parseInt(argv(4));

	test(new Histogram.Serial(), inputSize, histoSize, maxValue, maxAsyncs, numTrials);
    }
}
